from functools import cached_property

import numpy as np
from astropy import units as u
from astropy.modeling.fitting import LMLSQFitter, LinearLSQFitter
from astropy.modeling.models import Linear1D
from astropy.table import QTable, hstack
from gwcs import coordinate_frames as cf
from gwcs import wcs

from specreduce.compat import Spectrum

__all__ = [
    'WavelengthCalibration1D'
]


def _check_arr_monotonic(arr):
    # returns True if ``arr`` is either strictly increasing or strictly
    # decreasing, otherwise returns False.

    sorted_increasing = np.all(arr[1:] >= arr[:-1])
    sorted_decreasing = np.all(arr[1:] <= arr[:-1])
    return sorted_increasing or sorted_decreasing


class WavelengthCalibration1D():

    def __init__(self, input_spectrum, matched_line_list=None, line_pixels=None,
                 line_wavelengths=None, catalog=None, input_model=Linear1D(), fitter=None):
        """
        input_spectrum: `~specutils.Spectrum1D`
            A one-dimensional Spectrum calibration spectrum from an arc lamp or similar.
        matched_line_list: `~astropy.table.QTable`, optional
            An `~astropy.table.QTable` table with (minimally) columns named
            "pixel_center" and "wavelength" with known corresponding line pixel centers
            and wavelengths populated.
        line_pixels: list, array, `~astropy.table.QTable`, optional
            List or array of line pixel locations to anchor the wavelength solution fit.
            Can also be input as an `~astropy.table.QTable` table with (minimally) a column
            named "pixel_center".
        line_wavelengths: `~astropy.units.Quantity`, `~astropy.table.QTable`, optional
            `astropy.units.Quantity` array of line wavelength values corresponding to the
            line pixels defined in ``line_list``, assumed to be in the same
            order. Can also be input as an `~astropy.table.QTable` with (minimally)
            a "wavelength" column.
        catalog: list, str, `~astropy.table.QTable`, optional
            The name of a catalog of line wavelengths to load and use in automated and
            template-matching line matching. NOTE: This option is currently not implemented.
        input_model: `~astropy.modeling.Model`
            The model to fit for the wavelength solution. Defaults to a linear model.
        fitter: `~astropy.modeling.fitting.Fitter`, optional
            The fitter to use in optimizing the model fit. Defaults to
            `~astropy.modeling.fitting.LinearLSQFitter` if the model to fit is linear
            or `~astropy.modeling.fitting.LMLSQFitter` if the model to fit is non-linear.

        Note that either ``matched_line_list`` or ``line_pixels`` must be specified,
        and if ``matched_line_list`` is not input, at least one of ``line_wavelengths``
        or ``catalog`` must be specified.
        """
        self._input_spectrum = input_spectrum
        self._input_model = input_model
        self._cached_properties = ['fitted_model', 'residuals', 'wcs']
        self.fitter = fitter
        self._potential_wavelengths = None
        self._catalog = catalog

        if not isinstance(input_spectrum, Spectrum):
            raise ValueError('Input spectrum must be Spectrum.')

        # We use either line_pixels or matched_line_list to create self._matched_line_list,
        # and check that various requirements are fulfilled by the input args.
        if matched_line_list is not None:
            pixel_arg = "matched_line_list"
            if not isinstance(matched_line_list, QTable):
                raise ValueError("matched_line_list must be an astropy.table.QTable.")
            self._matched_line_list = matched_line_list
        elif line_pixels is not None:
            pixel_arg = "line_pixels"
            if isinstance(line_pixels, (list, np.ndarray)):
                self._matched_line_list = QTable([line_pixels], names=["pixel_center"])
            elif isinstance(line_pixels, QTable):
                self._matched_line_list = line_pixels
        else:
            raise ValueError("Either matched_line_list or line_pixels must be specified.")

        if "pixel_center" not in self._matched_line_list.columns:
            raise ValueError(f"{pixel_arg} must have a 'pixel_center' column.")

        if self._matched_line_list["pixel_center"].unit is None:
            self._matched_line_list["pixel_center"].unit = u.pix

        # check that pixels are monotonic
        if not _check_arr_monotonic(self._matched_line_list["pixel_center"]):
            raise ValueError('Pixels must be strictly increasing or decreasing.')

        # now that pixels have been determined from input, figure out wavelengths.
        if (line_wavelengths is None and catalog is None
                and "wavelength" not in self._matched_line_list.columns):
            raise ValueError("You must specify at least one of line_wavelengths, "
                             "catalog, or 'wavelength' column in matched_line_list.")

        # Sanity checks on line_wavelengths value
        if line_wavelengths is not None:
            if (isinstance(self._matched_line_list, QTable) and
                    "wavelength" in self._matched_line_list.columns):
                raise ValueError("Cannot specify line_wavelengths separately if there is"
                                 " a 'wavelength' column in matched_line_list.")
            if len(line_wavelengths) != len(self._matched_line_list):
                raise ValueError("If line_wavelengths is specified, it must have the same "
                                 f"length as {pixel_arg}")
            if not isinstance(line_wavelengths, (u.Quantity, QTable)):
                raise ValueError("line_wavelengths must be specified as an astropy.units.Quantity"
                                 " array or as an astropy.table.QTable")

            # make sure wavelengths (or freq) are monotonic and add wavelengths
            # to _matched_line_list
            if isinstance(line_wavelengths, u.Quantity):
                if not _check_arr_monotonic(line_wavelengths):
                    if str(line_wavelengths.unit.physical_type) == "frequency":
                        raise ValueError('Frequencies must be strictly increasing or decreasing.')
                    raise ValueError('Wavelengths must be strictly increasing or decreasing.')

                self._matched_line_list["wavelength"] = line_wavelengths

            elif isinstance(line_wavelengths, QTable):
                if not _check_arr_monotonic(line_wavelengths['wavelength']):
                    raise ValueError('Wavelengths must be strictly increasing or decreasing.')
                self._matched_line_list = hstack([self._matched_line_list, line_wavelengths])

        # Parse desired catalogs of lines for matching.
        if catalog is not None:
            # For now we avoid going into the later logic and just throw an error
            raise NotImplementedError("No catalogs are available yet, please input "
                                      "wavelengths with line_wavelengths or as a "
                                      f"column in {pixel_arg}")

            if isinstance(catalog, QTable):
                if "wavelength" not in catalog.columns:
                    raise ValueError("Catalog table must have a 'wavelength' column.")
                self._catalog = catalog
            else:
                # This will need to be updated to match up with Tim's catalog code
                if isinstance(catalog, list):
                    self._catalog = catalog
                else:
                    self._catalog = [catalog]
                for cat in self._catalog:
                    if isinstance(cat, str):
                        if cat not in self._available_catalogs:
                            raise ValueError(f"Line list '{cat}' is not an available catalog.")

    def identify_lines(self):
        """
        ToDo: Code matching algorithm between line pixel locations and potential line
        wavelengths from catalogs.
        """
        pass

    def _clear_cache(self, *attrs):
        """
        provide convenience function to clearing the cache for cached_properties
        """
        if not len(attrs):
            attrs = self._cached_properties
        for attr in attrs:
            if attr in self.__dict__:
                del self.__dict__[attr]

    @property
    def available_catalogs(self):
        return self._available_catalogs

    @property
    def input_spectrum(self):
        return self._input_spectrum

    @input_spectrum.setter
    def input_spectrum(self, new_spectrum):
        # We want to clear the refined locations if a new calibration spectrum is provided
        self._clear_cache()
        self._input_spectrum = new_spectrum

    @property
    def input_model(self):
        return self._input_model

    @input_model.setter
    def input_model(self, input_model):
        self._clear_cache()
        self._input_model = input_model

    @cached_property
    def fitted_model(self):
        # computes and returns WCS after fitting self.model to self.refined_pixels
        x = self._matched_line_list["pixel_center"]
        y = self._matched_line_list["wavelength"]

        if self.fitter is None:
            # Flexible defaulting if self.fitter is None
            if self.input_model.linear:
                fitter = LinearLSQFitter(calc_uncertainties=True)
            else:
                fitter = LMLSQFitter(calc_uncertainties=True)
        else:
            fitter = self.fitter

        # Fit the model
        return fitter(self.input_model, x, y)

    @cached_property
    def residuals(self):
        """
        calculate fit residuals between matched line list pixel centers and
        wavelengths and the evaluated fit model.
        """

        x = self._matched_line_list["pixel_center"]
        y = self._matched_line_list["wavelength"]

        # Get the fit residuals by evaulating model
        return y - self.fitted_model(x)

    @cached_property
    def wcs(self):
        # Build a GWCS pipeline from the fitted model
        pixel_frame = cf.CoordinateFrame(1, "SPECTRAL", [0,], axes_names=["x",], unit=[u.pix,])
        spectral_frame = cf.SpectralFrame(axes_names=["wavelength",],
                                          unit=[self._matched_line_list["wavelength"].unit,])

        pipeline = [(pixel_frame, self.fitted_model), (spectral_frame, None)]

        wcsobj = wcs.WCS(pipeline)

        return wcsobj

    def apply_to_spectrum(self, spectrum=None):
        # returns Spectrum with wavelength calibration applied
        # actual line refinement and WCS solution should already be done so that this can
        # be called on multiple science sources
        spectrum = self.input_spectrum if spectrum is None else spectrum
        updated_spectrum = Spectrum(spectrum.flux, wcs=self.wcs, mask=spectrum.mask,
                                    uncertainty=spectrum.uncertainty)
        return updated_spectrum



import astropy.units as u
import numpy as np
import pytest
from astropy.modeling import models
from astropy.modeling.polynomial import Polynomial1D
from astropy.nddata import StdDevUncertainty
from gwcs import wcs
from specreduce.wavesol1d import _diff_poly1d, WavelengthSolution1D
from specreduce.compat import Spectrum


ref_pixel = 250.0
p2w = models.Shift(ref_pixel) | models.Polynomial1D(degree=3, c0=1, c1=0.2, c2=0.001)
pix_bounds = (0, 500)
wav_bounds = p2w(pix_bounds)


@pytest.fixture
def mk_ws_without_transform():
    return WavelengthSolution1D(None, pix_bounds, u.angstrom)


@pytest.fixture
def mk_ws_with_transform():
    return WavelengthSolution1D(p2w, pix_bounds, u.angstrom)


@pytest.fixture
def mk_spectrum():
    return Spectrum(
        flux=np.ones(pix_bounds[1]) * u.DN,
        spectral_axis=np.arange(pix_bounds[1]) * u.pix,
        uncertainty=StdDevUncertainty(np.ones(pix_bounds[1])),
    )


def test_diff_poly1d():
    p = _diff_poly1d(Polynomial1D(3, c0=1.0, c1=2.0, c2=3.0, c3=4.0))
    np.testing.assert_array_equal(p.parameters, [2.0, 6.0, 12.0])


def test_init():
    ws = WavelengthSolution1D(p2w, pix_bounds, u.angstrom)
    assert ws._p2w is p2w
    assert ws.bounds_pix == pix_bounds
    assert ws.unit == u.angstrom
    assert "w2p" not in ws.__dict__
    assert "p2d_dldx" not in ws.__dict__
    assert "gwcs" not in ws.__dict__

    # Test that the cached properties are created correctly
    ws.w2p(0.5)
    assert "w2p" in ws.__dict__
    ws.p2w_dldx(pix_bounds[0])
    assert "p2w_dldx" in ws.__dict__
    wcs = ws.gwcs  # noqa: F841
    assert "gwcs" in ws.__dict__

    # Test that the cached properties are deleted correctly
    ws.p2w = p2w
    assert "w2p" not in ws.__dict__
    assert "p2d_dldx" not in ws.__dict__
    assert "gwcs" not in ws.__dict__

    ws = WavelengthSolution1D(p2w, pix_bounds, u.micron)
    assert ws.unit == u.micron

    ws = WavelengthSolution1D(None, pix_bounds, u.angstrom)
    assert ws._p2w is None


def test_resample(mk_spectrum, mk_ws_with_transform, mk_ws_without_transform):
    ws = mk_ws_with_transform
    spectrum = mk_spectrum

    # Resample a spectrum with uncertainty
    resampled = ws.resample(spectrum, nbins=50)
    assert resampled is not None
    assert len(resampled.flux) == 50
    assert resampled.flux.unit == u.DN / u.angstrom

    pix_edges = np.arange(spectrum.spectral_axis.size + 1) - 0.5
    f0 = (spectrum.flux.value * np.diff(ws._p2w(pix_edges))).sum()
    f1 = (resampled.flux.value * np.diff(resampled.spectral_axis.value)[0]).sum()
    np.testing.assert_approx_equal(f0, f1, 5)

    resampled = ws.resample(spectrum, wlbounds=wav_bounds)
    resampled = ws.resample(spectrum, bin_edges=np.linspace(*wav_bounds, num=50))

    # Resample a spectrum without uncertainty
    spectrum.uncertainty = None
    resampled = ws.resample(spectrum, nbins=50)
    assert resampled.uncertainty is not None

    ws = mk_ws_without_transform
    with pytest.raises(ValueError, match="Wavelength solution not set."):
        ws.resample(mk_spectrum)

    ws = mk_ws_with_transform
    with pytest.raises(ValueError, match="Number of bins must be non-zero and positive"):
        ws.resample(mk_spectrum, nbins=-5)


def test_pix_to_wav(mk_ws_with_transform):
    ws = mk_ws_with_transform
    pix = np.array([1, 2, 3, 4, 5])
    np.testing.assert_array_equal(ws.pix_to_wav(pix), p2w(pix))

    pix = np.ma.masked_array([1, 2, 3], mask=[0, 1, 0])
    wav = ws.pix_to_wav(pix)
    np.testing.assert_array_equal(wav.data, p2w(pix.data))
    np.testing.assert_array_equal(wav.mask, np.array([0, 1, 0]))


def test_wav_to_pix(mk_ws_with_transform):
    ws = mk_ws_with_transform
    pix_values_orig = np.array([1, 2, 3, 4, 5])
    pix_values_tran = ws.wav_to_pix(ws.pix_to_wav(pix_values_orig))
    np.testing.assert_array_almost_equal(pix_values_orig, pix_values_tran)

    pix_values_orig = np.ma.masked_array([1, 2, 3, 4, 5], mask=[0, 1, 0, 1, 0])
    pix_values_tran = ws.wav_to_pix(ws.pix_to_wav(pix_values_orig))
    np.testing.assert_array_almost_equal(pix_values_orig.data, pix_values_tran.data)
    np.testing.assert_array_almost_equal(pix_values_orig.mask, pix_values_tran.mask)


def test_wcs_creates_valid_gwcs_object(mk_ws_with_transform):
    wc = mk_ws_with_transform
    wcs_obj = wc.gwcs
    assert wcs_obj is not None
    assert isinstance(wcs_obj, wcs.WCS)
    assert wcs_obj.output_frame.unit[0] == u.angstrom



import numpy as np
import pytest
from astropy.modeling import fitting, models
from astropy.nddata import NDData
import astropy.units as u
from specreduce.utils.synth_data import make_2d_trace_image
from specreduce.tracing import Trace, FlatTrace, ArrayTrace, FitTrace

IM = make_2d_trace_image()


def mk_img(nrows=200, ncols=160, nan_slices=None, add_noise=True):
    """
    Makes a gaussian image for testing, with optional added gaussian
    nosie and optional data values set to NaN.
    """

    # NOTE: Will move this to a fixture at some point.

    sigma_pix = 4
    col_model = models.Gaussian1D(amplitude=1, mean=nrows / 2, stddev=sigma_pix)
    noise = 0
    if add_noise:
        np.random.seed(7)
        sigma_noise = 1
        noise = np.random.normal(scale=sigma_noise, size=(nrows, ncols))

    index_arr = np.tile(np.arange(nrows), (ncols, 1))
    img = col_model(index_arr.T) + noise

    if nan_slices:  # add nans in data
        for s in nan_slices:
            img[s] = np.nan

    return img * u.DN


# test basic trace class
def test_basic_trace():
    t_pos = IM.shape[0] / 2
    t = Trace(IM)

    assert t[0] == t_pos
    assert t[0] == t[-1]
    assert t.shape[0] == IM.shape[1]

    t.shift(100)
    assert t[0] == 600.0

    t.shift(-1000)
    assert np.ma.is_masked(t[0])


# test flat traces
def test_flat_trace():
    t = FlatTrace(IM, 550.0)

    assert t.trace_pos == 550
    assert t[0] == 550.0
    assert t[0] == t[-1]

    t.set_position(400.0)
    assert t[0] == 400.0


def test_negative_flat_trace_err():
    # make sure correct error is raised when trying to create FlatTrace with
    # negative trace_pos
    with pytest.raises(ValueError, match="must be positive."):
        FlatTrace(IM, trace_pos=-1)
    with pytest.raises(ValueError, match="must be positive."):
        FlatTrace(IM, trace_pos=0)


# test array traces
def test_array_trace():
    arr = np.ones_like(IM[0]) * 550.0
    t = ArrayTrace(IM, arr)

    assert t[0] == 550.0
    assert t[0] == t[-1]

    t.shift(100)
    assert t[0] == 650.0

    t.shift(-1000)
    assert np.ma.is_masked(t[0])

    arr_long = np.ones(5000) * 550.0
    t_long = ArrayTrace(IM, arr_long)

    assert t_long.shape[0] == IM.shape[1]

    arr_short = np.ones(50) * 550.0
    t_short = ArrayTrace(IM, arr_short)

    assert t_short[0] == 550.0
    assert np.ma.is_masked(t_short[-1])
    assert t_short.shape[0] == IM.shape[1]

    # make sure nonfinite data in input `trace` is masked
    arr[0:5] = np.nan
    t = ArrayTrace(IM, arr)
    assert np.all(t.trace.mask[0:5])
    assert np.all(t.trace.mask[5:] == 0)


# test fitted traces
@pytest.mark.filterwarnings("ignore:Model is linear in parameters")
def test_fit_trace():
    # create image (process adapted from compare_extractions.ipynb)
    np.random.seed(7)
    nrows = 200
    ncols = 160
    sigma_pix = 4
    sigma_noise = 1

    col_model = models.Gaussian1D(amplitude=1, mean=nrows / 2, stddev=sigma_pix)
    noise = np.random.normal(scale=sigma_noise, size=(nrows, ncols))

    index_arr = np.tile(np.arange(nrows), (ncols, 1))
    img = col_model(index_arr.T) + noise

    # calculate trace on normal image
    t = FitTrace(img, bins=20)

    # test shifting
    shift_up = int(-img.shape[0] / 4)
    t_shift_up = t.trace + shift_up

    shift_out = img.shape[0]

    t.shift(shift_up)
    assert np.sum(t.trace == t_shift_up) == t.trace.size, "valid shift failed"

    t.shift(shift_out)
    assert t.trace.mask.all(), "invalid values not masked"

    # test peak_method and trace_model options
    tg = FitTrace(img, bins=20, peak_method="gaussian", trace_model=models.Legendre1D(3))
    tc = FitTrace(img, bins=20, peak_method="centroid", trace_model=models.Chebyshev1D(2))
    tm = FitTrace(img, bins=20, peak_method="max", trace_model=models.Spline1D(degree=3))
    # traces should all be close to 100
    # (values may need to be updated on changes to seed, noise, etc.)
    assert np.max(abs(tg.trace - 100)) < sigma_pix
    assert np.max(abs(tc.trace - 100)) < 3 * sigma_pix
    assert np.max(abs(tm.trace - 100)) < 6 * sigma_pix
    with pytest.raises(ValueError):
        t = FitTrace(img, peak_method="invalid")

    window = 10
    guess = int(nrows / 2)
    img_win_nans = img.copy()
    img_win_nans[guess - window : guess + window] = np.nan

    # ensure float bin values trigger a warning but no issues otherwise
    with pytest.warns(UserWarning, match="TRACE: Converting bins to int"):
        FitTrace(img, bins=20.0, trace_model=models.Polynomial1D(2))

    # ensure non-equipped models are rejected
    with pytest.raises(ValueError, match=r"trace_model must be one of"):
        FitTrace(img, trace_model=models.Hermite1D(3))

    # ensure a bin number below 4 is rejected
    with pytest.raises(ValueError, match="bins must be >= 4"):
        FitTrace(img, bins=3)

    # ensure a bin number below degree of trace model is rejected
    with pytest.raises(ValueError, match="bins must be > "):
        FitTrace(img, bins=4, trace_model=models.Chebyshev1D(5))

    # ensure number of bins greater than number of dispersion pixels is rejected
    with pytest.raises(ValueError, match=r"bins must be <"):
        FitTrace(img, bins=ncols + 1)


def test_fit_trace_gaussian_all_zero():
    """
    Test fit_trace when peak_method is 'gaussian', which uses DogBoxLSQFitter
    for the fit for each bin peak and does not work well with all-zero columns.
    In this case, an all zero bin should fall back to NaN to for its'
    peak to be filtered out in the final fit for the trace.
    """
    img = mk_img(ncols=100)
    # add some all-zero columns so there is an all-zero bin
    img[:, 10:20] = 0

    t = FitTrace(img, bins=10, peak_method='gaussian')

    # this is a pretty flat trace, so make sure the fit reflects that
    assert np.all((t.trace >= 99) & (t.trace <= 101))


@pytest.mark.filterwarnings("ignore:The fit may be unsuccessful")
@pytest.mark.filterwarnings("ignore:Model is linear in parameters")
class TestMasksTracing:
    """
    There are three currently implemented options for masking in FitTrace: filter,
    omit, and zero_fill. Trace, FlatTrace, and ArrayTrace do not have
    `mask_treatment` options as input because masked/nonfinite values in the data
    are not relevant for those trace types as they are not affected by masked
    input data. The tests in this class test masking options for FitTrace, as
    well as some basic tests (errors, etc) for the other trace types.
    """

    def test_flat_and_basic_trace_mask(self):
        """
        Mask handling is not relevant for basic and flat trace - nans or masked
        values in the input image will not impact the trace value. The attribute
        should be initialized though, and be one of the valid options ([None]
        in this case), for consistancy with all other Specreduce operations.
        Note that unlike FitTrace, a fully-masked image should NOT result in an
        error raised because the trace does not depend on the data.
        """

        img = mk_img(nrows=5, ncols=5)

        basic_trace = Trace(img)
        assert basic_trace.mask_treatment is None

        flat_trace = FlatTrace(img, trace_pos=2)
        assert flat_trace.mask_treatment is None

        arr = [1, 2, np.nan, 3, 4]
        array_trace = ArrayTrace(img, arr)
        assert array_trace.mask_treatment is None

    def test_array_trace_masking(self):
        """
        The `trace` input to ArrayTrace can be a masked array, or an array
        containing nonfinite data which will be converted to a masked array.
        Additionally, if any padding needs to be added, the returned trace will
        be a masked array. Otherwise, it should be a regular array.

        Even though an ArrayTrace may have nans or masked values
        in the input 1D array for the trace, `mask_treatment_method` refers
        to how masked values in the input data should be treated. Nans / masked
        values passed in the array trace should be considered intentional, so
        also test that `mask_treatment` is initialized to None.
        """
        img = mk_img(nrows=10, ncols=10)

        # non-finite data in input trace should be masked out
        trace_arr = np.array((1, 2, np.nan, 4, 5))
        array_trace = ArrayTrace(img, trace_arr)
        assert array_trace.trace.mask[2]
        assert isinstance(array_trace.trace, np.ma.MaskedArray)

        # and combined with any input masked data
        trace_arr = np.ma.MaskedArray([1, 2, np.nan, 4, 5], mask=[1, 0, 0, 0, 0])
        array_trace = ArrayTrace(img, trace_arr)
        assert array_trace.trace.mask[0]
        assert array_trace.trace.mask[2]
        assert isinstance(array_trace.trace, np.ma.MaskedArray)

        # check that mask_treatment is None as there are no valid choices
        assert array_trace.mask_treatment is None

        # check that if array is fully finite and not masked, that the returned
        # trace is a normal array, not a masked array
        trace = ArrayTrace(img, np.ones(100))
        assert isinstance(trace.trace, np.ndarray)
        assert not isinstance(trace.trace, np.ma.MaskedArray)

        # ensure correct warning is raised when entire trace is masked.
        trace_arr = np.ma.MaskedArray([1, 2, np.nan, 4, 5], mask=[1, 1, 0, 1, 1])
        with pytest.warns(UserWarning, match=r"Entire trace array is masked."):
            array_trace = ArrayTrace(img, trace_arr)

    def test_fit_trace_fully_masked_image(self):
        """
        Test that the correct warning is raised when a fully masked image is
        encountered. Also test that when a non-fully masked image is provided,
        but `window` is set and the image is fully masked within that window,
        that the correct error is raised.
        """

        # make simple gaussian image.
        img = mk_img()

        # create same-shaped variations of image with nans in data array
        # which will be masked within FitTrace.
        nrows = 200
        ncols = 160
        img_all_nans = np.tile(np.nan, (nrows, ncols))

        # error on trace of all-nan image
        with pytest.raises(ValueError, match=r"Image is fully masked. Check for invalid values."):
            FitTrace(img_all_nans)

        window = 10
        guess = int(nrows / 2)
        img_win_nans = img.copy()
        img_win_nans[guess - window : guess + window] = np.nan

        # error on trace of otherwise valid image with all-nan window around guess
        with pytest.raises(ValueError, match="pixels in window region are masked"):
            FitTrace(img_win_nans, guess=guess, window=window)

    def test_fit_trace_fully_masked_columns_warn_msg(self):
        """
        Test that the correct warning message is raised when fully masked columns
        (in a not-fully-masked image) are encountered in FitTrace. These columns
        will be set to NaN and filtered from the final all-bin fit (as tested in
        test_fit_trace_fully_masked_cols), but a warning message is raised.
        """
        img = mk_img()

        # test that warning (dependent on choice of `peak_method`) is raised when a
        # few bins are masked, and that theyre listed individually
        mask = np.zeros(img.shape, dtype=bool)
        mask[:, 100] = 1
        mask[:, 20] = 1
        mask[:, 30] = 1
        nddat = NDData(data=img, mask=mask, unit=u.DN)

        match_str = "All pixels in bins 20, 30, 100 are fully masked. Setting bin peaks to NaN."

        with pytest.warns(UserWarning, match=match_str):
            FitTrace(nddat, peak_method="max")

        with pytest.warns(UserWarning, match=match_str):
            FitTrace(nddat, peak_method="centroid")

        with pytest.warns(UserWarning, match=match_str):
            FitTrace(nddat, peak_method="gaussian")

        # and when many bins are masked, that the message is consolidated
        mask = np.zeros(img.shape, dtype=bool)
        mask[:, 0:21] = 1
        nddat = NDData(data=img, mask=mask, unit=u.DN)
        with pytest.warns(
            UserWarning,
            match="All pixels in bins "
            "0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ..., 20 are "
            "fully masked. Setting bin peaks to NaN.",
        ):
            FitTrace(nddat)

    @pytest.mark.filterwarnings("ignore:All pixels in bins")
    @pytest.mark.parametrize("mask_treatment", ["apply", "propagate", "apply_nan_only"])
    def test_fit_trace_fully_masked_cols(self, mask_treatment):
        """
        Create a test image with some fully-nan/masked columns, and test that
        when the final fit to all bin peaks is done for the trace, that these
        fully-masked columns are set to NaN and filtered during the final all-bin
        fit. Ignore the warning that is produced when this case is encountered (that
        is tested in `test_fit_trace_fully_masked_cols_warn_msg`.)
        """
        img = mk_img(nrows=10, ncols=11)

        # set some columns fully to nan, which will be masked out
        img[:, 7] = np.nan
        img[:, 4] = np.nan
        img[:, 0] = np.nan

        # also create an image that doesn't have nans in the data, but
        # is masked in the same locations, to make sure that is equivilant.

        # test peak_method = 'max'
        truth = [
            1.6346154,
            2.2371795,
            2.8397436,
            3.4423077,
            4.0448718,
            4.6474359,
            5.25,
            5.8525641,
            6.4551282,
            7.0576923,
            7.6602564,
        ]
        max_trace = FitTrace(img, peak_method="max", mask_treatment=mask_treatment)
        np.testing.assert_allclose(truth, max_trace.trace, atol=0.1)

        # peak_method = 'gaussian'
        truth = [
            1.947455,
            2.383634,
            2.8198131,
            3.2559921,
            3.6921712,
            4.1283502,
            4.5645293,
            5.0007083,
            5.4368874,
            5.8730665,
            6.3092455,
        ]
        max_trace = FitTrace(img, peak_method="gaussian", mask_treatment=mask_treatment)
        np.testing.assert_allclose(truth, max_trace.trace, atol=0.1)

        # peak_method = 'centroid'
        truth = [
            2.5318835,
            2.782069,
            3.0322546,
            3.2824402,
            3.5326257,
            3.7828113,
            4.0329969,
            4.2831824,
            4.533368,
            4.7835536,
            5.0337391,
        ]
        max_trace = FitTrace(img, peak_method="centroid", mask_treatment=mask_treatment)
        np.testing.assert_allclose(truth, max_trace.trace, atol=0.1)

    @pytest.mark.filterwarnings("ignore:All pixels in bins")
    @pytest.mark.parametrize(
        "peak_method,expected",
        [
            ("max", [5.0, 3.0, 5.0, 5.0, 7.0, 5.0, np.nan, 5.0, 5.0, 5.0, 2.0, 5.0]),
            ("gaussian", [5.0, 5.0, 5.0, 5.0, 5.0, 5.0, np.nan, 5.0, 5.0, 5.0, 1.576, 5.0]),
            (
                "centroid",
                [
                    4.27108332,
                    2.24060342,
                    4.27108332,
                    4.27108332,
                    6.66827608,
                    4.27108332,
                    np.nan,
                    4.27108332,
                    4.27108332,
                    4.27108332,
                    1.19673467,
                    4.27108332,
                ],
            ),
        ],
    )
    def test_mask_treatment_apply(self, peak_method, expected):
        """
        Test for mask_treatment=apply for FitTrace.
        With this masking option, masked and non-finite data should be filtered
        when determining bin/column peak. Fully masked bins should be omitted
        from the final all-bin-peak fit for the Trace. Parametrized over different
        `peak_method` options.
        """

        # Make an image with some non-finite values.
        image1 = mk_img(
            nan_slices=[np.s_[4:8, 1:2], np.s_[2:7, 4:5], np.s_[:, 6:7], np.s_[3:9, 10:11]],
            nrows=10,
            ncols=12,
            add_noise=False,
        )

        # Also make an image that doesn't have nonf data values, but has masked
        # values at the same locations, to make sure they give the same results.
        mask = ~np.isfinite(image1)
        dat = mk_img(nrows=10, ncols=12, add_noise=False)
        image2 = NDData(dat, mask=mask)

        for imgg in [image1, image2]:
            # run FitTrace, with the testing-only flag _save_bin_peaks_testing set
            # to True to return the bin peak values before fitting the trace
            trace = FitTrace(imgg, peak_method=peak_method, _save_bin_peaks_testing=True)
            x_bins, y_bins = trace._bin_peaks
            np.testing.assert_allclose(y_bins, expected, atol=0.1)

            # check that final fit to all bins, accouting for fully-masked bins,
            # matches the trace
            fitter = fitting.LMLSQFitter()
            mask = np.isfinite(y_bins)
            all_bin_fit = fitter(trace.trace_model, x_bins[mask], y_bins[mask])
            all_bin_fit = all_bin_fit((np.arange(12)))

            np.testing.assert_allclose(trace.trace, all_bin_fit)

    @pytest.mark.filterwarnings("ignore:All pixels in bins")
    @pytest.mark.parametrize("peak_method", ["max", "gaussian", "centroid"])
    def test_mask_treatment_unsupported(self, peak_method):
        """
        Test to ensure the unsupported mask treatment methods for FitTrace
        raise a `ValueError`. Parametrized over different `peak_method` options.
        """

        image = mk_img(
            nan_slices=[np.s_[4:8, 1:2], np.s_[2:7, 4:5], np.s_[:, 6:7], np.s_[3:9, 10:11]],
            nrows=10,
            ncols=12,
            add_noise=False,
        )

        for method in "ignore", "zero_fill", "nan_fill", "apply_mask_only":
            with pytest.raises(ValueError):
                FitTrace(image, peak_method=peak_method, mask_treatment=method)

    @pytest.mark.filterwarnings("ignore:All pixels in bins")
    @pytest.mark.parametrize(
        "peak_method,expected",
        [
            ("max", [5.0, np.nan, 5.0, 5.0, np.nan, 5.0, np.nan, 5.0, 5.0, 5.0, np.nan, 5.0]),
            ("gaussian", [5.0, np.nan, 5.0, 5.0, np.nan, 5.0, np.nan, 5.0, 5.0, 5.0, np.nan, 5.0]),
            (
                "centroid",
                [
                    4.27108332,
                    np.nan,
                    4.27108332,
                    4.27108332,
                    np.nan,
                    4.27108332,
                    np.nan,
                    4.27108332,
                    4.27108332,
                    4.27108332,
                    np.nan,
                    4.27108332,
                ],
            ),
        ],
    )
    def test_mask_treatment_propagate(self, peak_method, expected):
        """
        Test for mask_treatment=`propagate` for FitTrace. Columns (assuming
        disp_axis==1) with any masked data values will be fully masked and
        therefore not contribute to the bin peaks. Parametrized over different
        `peak_method` options.
        """

        # Make an image with some non-finite values.
        image1 = mk_img(
            nan_slices=[np.s_[4:8, 1:2], np.s_[2:7, 4:5], np.s_[:, 6:7], np.s_[3:9, 10:11]],
            nrows=10,
            ncols=12,
            add_noise=False,
        )

        # Also make an image that doesn't have non-finite data values, but has masked
        # values at the same locations, to make sure those cases are equivalent
        mask = ~np.isfinite(image1)
        dat = mk_img(nrows=10, ncols=12, add_noise=False)
        image2 = NDData(dat, mask=mask)

        for imgg in [image1, image2]:

            # run FitTrace, with the testing-only flag _save_bin_peaks_testing set
            # to True to return the bin peak values before fitting the trace
            trace = FitTrace(
                imgg,
                peak_method=peak_method,
                mask_treatment="propagate",
                _save_bin_peaks_testing=True,
            )
            x_bins, y_bins = trace._bin_peaks
            np.testing.assert_allclose(y_bins, expected)

            # check that final fit to all bins, accouting for fully-masked bins,
            # matches the trace
            fitter = fitting.LevMarLSQFitter()
            mask = np.isfinite(y_bins)
            all_bin_fit = fitter(trace.trace_model, x_bins[mask], y_bins[mask])
            all_bin_fit = all_bin_fit((np.arange(12)))

            np.testing.assert_allclose(trace.trace, all_bin_fit)
